#' Cropping a figure
#'
#' Crops a geometry to the named input of extents
#'
#' @param geo sfc
#' @param bbox (named vector of optional xmin, xmax, ymin, ymax)
#'
#' @export
crop_geo <- function(geo, bbox) {
  # Getting a boundary box of the entire region and then setting the crop
  sf::st_agr(geo) <- "constant"
  sf::st_crop(geo, replace(sf::st_bbox(geo), names(bbox), bbox))
}

#' Builds a grid of cell center points
#'
#' The grid starts at (xmin, ymin) and is ordered by x (latitude) and then
#' by longitude. The grid defaults are PRCP_CU_GAUGE_V1.0CONUS_0.25deg.README
#' in https://ftp.cpc.ncep.noaa.gov/precip/CPC_UNI_PRCP/GAUGE_CONUS/DOCU/.
#'
#' "The CONUS field starts from the 0.25deg lat/lon grid box centering
#' at (129.875W; 20.125N), going from west to east first, and then south to
#' north."
#'
#' @param grid_size size of the grid in degrees
#' @param bbox named vector of the boundary box
#' @param crs the coordinate reference system
#' @return sf tibble of the grid center points
#' @export
grid_center_points <- function(grid_size = 0.25,
                               bbox = c("ymin"=20.125, "ymax"=49.875, "xmin"=-129.875, "xmax"=-55.125),
                               crs = "EPSG:4326") {

  center_points <- tidyr::crossing(
    lat = seq(bbox["ymin"], bbox["ymax"], by = grid_size),
    lon = seq(bbox["xmin"], bbox["xmax"], by = grid_size)) %>%
    dplyr::arrange(lat, lon) %>%
    dplyr::mutate(GRIDID = dplyr::row_number()) %>%
    sf::st_as_sf(coords = c("lon", "lat"), crs = crs)

  return(center_points)
}

#' Determines the between Polygons of the Geometry
#'
#' Determines the distance between the centroid of polygons. All pairs of
#' polygons are considered.
#'
#' @param geo sf tibble
#' @return tibble of geoid, adj_geoid, distance
#' @export
calc_distance_geo <- function(geo) {

  if (!("geoid" %in% colnames(geo))) {
    if ("GEOID" %in% colnames(geo)){
      rlang::warn("Using `GEOID` as `geoid`")
      geo <- geo %>%
        janitor::clean_names()
    } else{
      rlang::abort("`geo` must contain a column `geoid`")
    }
  }

  # Finding the centroids
  geo <- geo %>%
    dplyr::mutate(centroid = sf::st_centroid(geometry))
  centroids <- geo %>%
    sf::st_set_geometry(geo$centroid) %>%
    dplyr::select(geoid)

  # Computing the distance
  dist <- sf::st_distance(geo$centroid, by_element = FALSE)

  # Converting to a tibble
  rownames(dist) <- geo$geoid
  colnames(dist) <- geo$geoid

  tibble::as_tibble(dist) %>%
    dplyr::mutate(adj_geoid = geo$geoid) %>%
    tidyr::pivot_longer(!adj_geoid, names_to = "geoid", values_to = "distance") %>%
    dplyr::relocate(geoid)
}

#' Determines the Adjacent Polygons
#'
#' Determines which polygons are adjacent each other with the spdep::poly2nb
#' function
#'
#' @param geo sf tibble
#' @param ... additional arguments to pass to spdep::poly2nb
#' @param self_adj flag to indicate if self-adjacency should be considered
#' @return tibble of `geoid`, `adj_geoid` pairs
#' @seealso  spdep::poly2nb
#' @export
calc_adjacency <- function(geo, ..., self_adj = TRUE) {

  # Checking for case
  if (!("geoid" %in% colnames(geo))) {
    if ("GEOID" %in% colnames(geo)){
      rlang::warn("Using `GEOID` as `geoid`")
      geo <- geo %>%
        janitor::clean_names()
    } else{
      rlang::abort("`geo` must contain a column `geoid`")
    }
  }

  # Finding the adjacent neighbors (weights)
  W.nb <- spdep::poly2nb(geo, row.names = geo$geoid, ...)
  W <- spdep::nb2mat(W.nb, style = "B", zero.policy = TRUE)


  # Adding on an indentiy matrix (self-adjacent)
  if (self_adj) {
    W <- diag(x = 1, dim(W)) + W
  }

  # Converting to a tibble
  rownames(W) <- geo$geoid
  colnames(W) <- geo$geoid

  df <- tibble::as_tibble(W) %>%
    dplyr::mutate(adj_geoid = geo$geoid) %>%
    tidyr::pivot_longer(!adj_geoid, names_to = "geoid", values_to = "adjacent") %>%
    dplyr::relocate(geoid)

  df %>%
    dplyr::filter(adjacent == 1) %>%
    dplyr::select(-adjacent)
}

#' Calculates the inverse distance weight of a tibble
#'
#' Determines the inverse distance weight of a column in a tibble. The weight
#' is calculated as:
#'  1) Joining each of the geoid to it's neighbors and adding the value
#'  2) Determining the distance of the geoid to its neighbors
#'  3) Computing the weighted mean ( sum( dist ^ p * val) / sum(dist ^ p))
#'
#' When computing the weighted mean the identity (geoid == adj_geoid) is
#' removed.
#'
#' @param x input tibble. Must contain column provided in `col` and `geoid`
#' @param col <[`tidy-select`][dplyr_tidy_select]> of single column
#' @param p the distance weight (p = -1 for inverse distance)
#' @param n reserved
#' @param adj_list county adjacency list. Must have columns geoid and adj_geoid,
#' describing the county and it's adjacent neighbors
#' @param by vector of names to join by (keys) if geoid does not uniquely describe
#' the tibble
#' @param dist tibble of `geoid`, `adj_geoid` , and `distance`. If not
#' provided then the distance is determined from the supplied geometry.
#' Supplying the distance allows for caching the expensive distance results.
#' @export
#' @note If supplied a distance, shouldn't need to
distance_weight <- function(x, col, p = -1, n = 1, adj_list = NULL, by = NULL, dist = NULL) {
  col <- rlang::enquo(col)
  adj_col <- rlang::sym(glue::glue("adj_{rlang::as_label(col)}"))

  if (is.null(dist)) {
    checkmate::checkClass(x, c("sf"))
  }

  check_columns(x, c("geoid"))
  check_columns(adj_list, c("geoid", "adj_geoid"))
  n <- checkmate::checkCount(n, positive = TRUE)
  if (n != 1) {
    rlang::abort("Sorry, only n = 1 is implemented!")
  }

  if (is.null(by)) {
    by <- c("geoid", rlang::as_name(col))
    if (nrow(x) != length(unique(x$geoid))){
        rlang::warn("geoid does not uniquely describe x! Consider using `by`.")
    }
  } else {
    # Make sure by doesn't include geoid, col
    by <- unique(c(c("geoid", rlang::as_name(col)), by))
  }

  # Creating a copy
  x_orig <- x
  x <- x_orig %>%
    dplyr::select(tidyselect::all_of(by))


  if (is.null(adj_list)) {
    # Computing the Adjacency
    rlang::warn("Computing the adjacency is in beta! May differ from expected!")
    adj_list <- calc_adjacency(x)
  }

  # Computing the distance
  if (is.null(dist)) {
    dist <- calc_distance(x)
  }
  x <- sf::st_drop_geometry(x)

  # Joining on the adj_list
  x_adj <- dplyr::left_join(
    x,
    adj_list,
    by = c("geoid" = "geoid")
  )

  x_adj <- dplyr::left_join(
    x_adj,
    x %>% dplyr::rename(!!adj_col := {{ col }}),
    by = c("adj_geoid" = "geoid")
  )

  x_adj <- dplyr::left_join(
    x_adj,
    dist,
    by = c("geoid" = "geoid", "adj_geoid" = "adj_geoid")
  )

  # Inverse Weight
  x_add <- x_adj %>%
    dplyr::mutate(distance = units::drop_units(distance)) %>%
    dplyr::group_by(geoid) %>%
    dplyr::filter(adj_geoid != geoid) %>%
    dplyr::summarise("{{col}}_weighted_avg" := sum(distance^p * !!adj_col, na.rm = TRUE) / sum(distance^p, na.rm = TRUE), .groups = "drop")

  # Adding them back on
  by <- by[by != rlang::as_name(col)]
  x_res <- dplyr::full_join(
    x_orig,
    x_add,
    by = by
  )
  return(x_res)
}

#' Determines the Intersection of each grid to the geo
#'
#' Determines the area that each grid cell intersects with the supplied
#' geometry
#' @param geo sf of the polygon. Must include column geoid and a geometry
#' @param grid sf of the grid
#' @return sf tibble of the grid regions that intersect each region of the
#' polygon. The fraction of each grid region in
#' @export
grid_geo_intersection <- function(geo, grid) {

  checkmate::checkClass(geo, "sf")
  checkmate::checkClass(grid, "sf")

  # Transform grid to crs of geo
  grid <- sf::st_transform(grid, sf::st_crs(geo))

  # Filtering to make problem smaller
  gg_intersection <- sf::st_intersection(grid, geo)

  # Computing the area, and then the fraction
  gg_intersection <- gg_intersection %>%
    dplyr::mutate(geometry = sf::st_make_valid(geometry)) %>%
    dplyr::mutate(area = sf::st_area(geometry)) %>%
    dplyr::group_by(geoid) %>%
    dplyr::mutate(fraction = units::drop_units(area / sum(area))) %>%
    dplyr::ungroup()

  return(gg_intersection)

}
