#' Builds Convex Hulls from Cyclone Tracks
#'
#' Builds buffers and convex hulls from cyclone tracks. The buffers are created
#' using the `.r` parameter as the radius, and the convex hull is created for
#' each group specified in `grp_vars`, after it has been filtered by the
#' arguments provided.
#'
#' Filter is completed after the data is grouped, run length encoded, and
#' re-grouped. This ensures that each group of the ibtracs data is continuous.
#'
#' @param data ibtracs dataset
#' @param r radius variable to use for the buffers
#' @param grp_vars  <[`tidy-select`][dplyr_tidy_select]>  of the variables to group on. These are run
#' length encoded (sorted by usa_atcfid and iso_time) and used for the grouping to
#' ensure that the track order is kept
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]> filter arguments. These are
#' applied after the run length encoding to preserve the order / ensure that the
#' ibtracs data is continuous in time.
#' @param .summarize flag to determine if summarized result should be returned
#' @param .parallel flag to determine if `multidplyr` parallel processing should
#' be used
#'
#' @return sf tibble of the `grp_vars`, summerized (`sf::st_union`) and the
#' first and last time-step encapsulated in the summerized data. The
#' intersections with `.geo` are provided if `.geo` is not null.
#' @export
#'
build_hulls <- function(data, r, grp_vars, ...,
                        .summarize = FALSE, .parallel = FALSE) {
  hulls <- NULL
  args <- rlang::enexprs(...)

  # Marking the rows to keep
  to_keep <- data %>%
    sf::st_drop_geometry() %>%
    dplyr::group_by(dplyr::across({{ grp_vars }})) %>%
    dplyr::filter(!!!args) %>%
    dplyr::mutate(to_keep = TRUE)

  data <- data %>%
    dplyr::left_join(y = to_keep, by = colnames(sf::st_drop_geometry(data)))

  # Adding the run length encoding
  data <- data %>%
    dplyr::group_by(dplyr::across({{ grp_vars }})) %>%
    dplyr::arrange(iso_time, .by_group = TRUE) %>%
    dplyr::ungroup() %>%
    tidyr::unite(rle_var, {{ grp_vars }}, to_keep,  remove = FALSE) %>%
    dplyr::mutate(rle = rle_encode(rle_var))

  # Grouping the data by the provided and then rle
  grps <- data %>%
    dplyr::group_by(dplyr::across({{ grp_vars }})) %>%
    dplyr::group_by(rle, .add = TRUE) %>%
    dplyr::arrange(iso_time, .by_group = TRUE)

  # Filtering
  grps <- grps %>%
    dplyr::filter(!is.na({{ r }})) %>%
    dplyr::filter(!!!args)


  if (nrow(grps) == 0) {
    keys <- dplyr::group_keys(grps) %>%
      dplyr::select(-rle)
    data <- keys %>%
      dplyr::mutate(hull = sf::st_as_sfc("POLYGON(EMPTY)")) %>%
      sf::st_as_sf()
    rlang::warn("Did not build any hulls! Empty Geometry Returned!")
    return(data)
  }

  # warng grps for n() > 1
  warn_group <- function(tbl, grp, ...) {
    if (nrow(tbl) == 1) {
      rlang::warn(glue::glue("Group {grp} only has one entry!"))
    }
    if (nrow(tbl) == 0) {
      rlang::warn(glue::glue("Group {grp} only no data!"))
    }
  }
  dplyr::group_walk(grps, .f = warn_group)


  if (.parallel) {
    if (sf::st_crs(ibtracs) != sf::st_crs("EPSG:4326")) {
      rlang::warn("Using an EPSG that ios not 4326 may result in the code failing! Use .parallel = FALSE.")
    }

    # Cluster Setup
    max_cores <- parallel::detectCores()
    cores <- max(1, max_cores - 2)
    cluster <- multidplyr::new_cluster(cores)

    # Distributing the groups
    shards <- grps %>%
      multidplyr::partition(cluster)

    # Doing the work
    hulls <- shards %>%
      # Creating the buffer
      dplyr::mutate(buffer = sf::st_buffer(geometry, dist = {{ r }}, nQuadSegs = 180)) %>%
      dplyr::mutate(buffer = sf::st_make_valid(buffer)) %>%
      # Creating the convex hull
      dplyr::mutate(lag_buffer = dplyr::lag(buffer, order_by = iso_time)) %>%
      dplyr::mutate(track = dplyr::case_when(
        !is.na(lag_buffer) ~ sf::st_union(buffer, lag_buffer),
        is.na(lag_buffer) ~ buffer
        )
      ) %>%
      dplyr::mutate(hull = sf::st_convex_hull(track))

    if (.summarize) {
      hulls <- hulls %>%
        dplyr::summarise(
          start = dplyr::first(iso_time),
          end = dplyr::last(iso_time),
          geometry = sf::st_union(hull) %>% sf::st_make_valid()
        )
    } else {
      hulls <- hulls %>%
        dplyr::mutate(hull = sf::st_make_valid(hull)) %>%
        dplyr::select(tidyselect::all_of(dplyr::group_vars(grps)), iso_time, {{ r }}, geometry, hull)

      # Renaming and setting geometry columns
      hulls <- hulls %>%
        dplyr::rename(point = geometry, geometry = hull)
    }

    # Collecting and cleanup
    hulls <- hulls %>%
      dplyr::collect() %>%
      dplyr::ungroup()

    # Ensuring an sf
    hulls <- sf::st_as_sf(hulls)

    rm(cluster)
  } else {
    hulls <- grps %>%
      # Creating the buffer
      dplyr::mutate(buffer = sf::st_buffer(geometry, dist = {{ r }})) %>%
      dplyr::mutate(buffer = sf::st_make_valid(buffer)) %>%
      # Creating the convex hull
      dplyr::mutate(lag_buffer = dplyr::lag(buffer)) %>%
      dplyr::rowwise() %>%
      dplyr::mutate(track = sf::st_union(buffer, lag_buffer)) %>%
      dplyr::mutate(hull = sf::st_convex_hull(track)) %>%
      dplyr::group_by(dplyr::across({{ grp_vars }})) %>%
      dplyr::group_by(rle, .add = TRUE)

    if (.summarize) {
      hulls <- hulls %>%
        dplyr::summarise(
          start = dplyr::first(iso_time),
          end = dplyr::last(iso_time),
          geometry = sf::st_union(hull) %>% sf::st_make_valid(),
          .groups = "drop"
        )
    } else {
      hulls <- hulls %>%
        dplyr::mutate(hull = sf::st_make_valid(hull)) %>%
        dplyr::select(tidyselect::all_of(dplyr::group_vars(grps)), iso_time, {{ r }}, geometry, hull)

      # Renaming and setting geometry columns
      hulls <- hulls %>%
        dplyr::rename(point = geometry, geometry = hull)
      sf::st_geometry(hulls) <- "geometry"
    }
  }

  # Removing the added variables
  hulls <- hulls %>%
    dplyr::ungroup() %>%
    dplyr::select(-dplyr::ends_with("rle"))

  # Promoting to SF
  if ("sf" %in% class(.data) & !("sf" %in% class(hulls))) {
    hulls <- sf::st_set_geometry(hulls, attr(.data, "sf_column"))
  }
  return(hulls)
}

#' Intersection of Hulls and Geometry
#'
#' Determines the intersection of the hulls and the geometry.
#'
#' @param data geometry of the hulls for intersection
#' @param .geo object of class sf. Must have a geometry column and `geoid`.
#' @param .parallel flag to determine if `multidplyr` parallel processing should
#' be used. If NULL, then parallel will be used if over 20 hulls
#' @export
hull_intersect <- function(.data, .geo, .parallel = NULL) {
  # Checking for parallel
  if (is.null(.parallel)) {
    .parallel <- nrow(.data) > 20
  }

  if (.parallel) {
    # Cluster Setup
    max_cores <- parallel::detectCores()
    cores <- max(1, max_cores - 2)
    cluster <- multidplyr::new_cluster(cores)
    multidplyr::cluster_copy(cluster, ".geo")

    # Distributing the groups
    shards <- data %>%
      multidplyr::partition(cluster)

    # Doing the work
    hulls <- shards %>%
      dplyr::mutate(
        geoid =
          purrr::map(geometry, ~ sf::st_filter(
            x = .geo, y = .x, .predicate = sf::st_intersects
          )$geoid)
      )

    # Collecting and cleanup
    hulls <- hulls %>%
      dplyr::collect()

    rm(cluster)
  } else {
    hulls <- data %>%
      dplyr::mutate(
        geoid =
          purrr::map(geometry, ~ sf::st_filter(
            x = .geo, y = .x, .predicate = sf::st_intersects
          )$geoid)
      )
  }
  return(hulls)
}
