#' Preprocessing of the IBTrACS Dataset
#'
#' Preprocessing the IBTrACS Dataset. Actions preformed are:
#' - Removes empty and constant columns and the following list of columns:
#'  `c(year, month, day, hour, usa_lat, usa_lon, wmo_agency, usa_agency,
#'     starts_with("mlc_"), starts_with('td6_), starts_with("ds8_),
#'     starts_with("tok_"), starts_with("cma_"), starts_with("hko_"),
#'     starts_with("new_"), starts_with("reu_"), starts_with("bom_"),
#'     starts_with("nad_"), starts_with("wel_"), starts_with("td5_"),
#'     starts_with("neu_"), starts_with("usa_sea"), usa_poci,
#'     usa_roci, usa_rmw, usa_eye, usa_gust)`
#'
#'
#' - Converts iso_time, wind radius, storm speed
#' - Compute maximum wind radius extents
#'
#' @param data ibtracs data
#' @return processed ibtracs data
#' @export
ibtracs_prepare <- function(data){
  radius_pattern <- "usa_r[0-9][0-9]_[n|s][e|w]"

  # Preliminary clean
  data <- data %>%
    janitor::clean_names() %>%
    dplyr::select(-tidyselect::any_of(c("year", "month", "day", "hour", "min"))) %>%
    dplyr::select(-tidyselect::any_of(c("usa_lat", "usa_lon"))) %>%
    dplyr::select(-tidyselect::any_of(c("wmo_agency", "usa_agency"))) %>%
    dplyr::select(
      -tidyselect::starts_with("mlc_"),
      -tidyselect::starts_with("td6_"),
      -tidyselect::starts_with("ds8_"),
      -tidyselect::starts_with("tok_"),
      -tidyselect::starts_with("cma_"),
      -tidyselect::starts_with("hko_"),
      -tidyselect::starts_with("new_"),
      -tidyselect::starts_with("reu_"),
      -tidyselect::starts_with("bom_"),
      -tidyselect::starts_with("nad_"),
      -tidyselect::starts_with("wel_"),
      -tidyselect::starts_with("td5_"),
      -tidyselect::starts_with("neu_"),
      -tidyselect::starts_with("usa_sea"),
      -usa_poci,
      -usa_roci,
      -usa_rmw,
      -usa_eye,
      -usa_gust
      )

  # Unit Conversion and Setting
  data <- data %>%
    dplyr::mutate(iso_time = lubridate::ymd_hms(iso_time, tz = "UTC")) %>%
    dplyr::mutate(dplyr::across(tidyselect::matches(radius_pattern), ~ units::set_units(.x, "nautical_mile"))) %>%
    dplyr::mutate(dplyr::across(tidyselect::matches(radius_pattern), ~ units::set_units(.x, "m"))) %>%
    dplyr::mutate(dplyr::across(c(dist2land, landfall), ~ .x * 1000)) %>%
    dplyr::mutate(dplyr::across(c(dist2land, landfall),  ~ units::set_units(.x, "m"))) %>%
    dplyr::mutate(storm_spd = units::set_units(storm_spd, "knots")) %>%
    dplyr::mutate(storm_spd = units::set_units(storm_spd, "m/s"))

  # Wind Extent Radius
  data <- data %>%
    dplyr::mutate(r_64 = pmax(usa_r64_ne, usa_r64_nw, usa_r64_se, usa_r64_sw, na.rm = TRUE)) %>%
    dplyr::mutate(r_50 = pmax(usa_r50_ne, usa_r50_nw, usa_r50_se, usa_r50_sw, na.rm = TRUE)) %>%
    dplyr::mutate(r_34 = pmax(usa_r34_ne, usa_r34_nw, usa_r34_se, usa_r34_sw, na.rm = TRUE)) %>%
    dplyr::mutate(
      bad_wind_extent_ts = dplyr::case_when(
        is.na(r_34) & usa_wind >=34 ~ 1,
        T ~ 0
        ),
      bad_wind_extent_hu = dplyr::case_when(
        is.na(r_64) & usa_wind >=64 ~ 1,
        T ~ 0
      )
    )

  # Gives warning about missing wind extents
  if(
    data %>%
    dplyr::tibble() %>%
    dplyr::filter(bad_wind_extent_ts == 1) %>%
    dplyr::slice_head(n=1) %>%
    nrow() >= 1
  )
    warning(
      paste0(
        data %>%
          dplyr::tibble() %>%
          dplyr::select(
            usa_atcfid,
            iso_time,
            name,
            season,
            bad_wind_extent_ts
          ) %>%
          dplyr::filter(bad_wind_extent_ts == 1) %>%
          dplyr::mutate(
            wind_extent_warning_id = name %>%
              stringr::str_to_title() %>%
              stringr::str_c(iso_time, sep = ", ")
          ) %>%
          dplyr::select(wind_extent_warning_id) %>%
          purrr::pluck() %>%
          unlist(use.names = F) %>%
          stringr::str_flatten_comma(", and "),
        " are Tropical Storm observation(s) with missing wind extents. These observation(s) are being dropped. Proceed with Caution."
      ) %>%
        paste0(sep="\n")
    )

  if(
    data %>%
    dplyr::tibble() %>%
    dplyr::filter(bad_wind_extent_hu == 1) %>%
    dplyr::slice_head(n=1) %>%
    nrow() >= 1
  )
    warning(
      paste0(
        data %>%
          dplyr::tibble() %>%
          dplyr::select(
            usa_atcfid,
            iso_time,
            name,
            season,
            bad_wind_extent_hu
          ) %>%
          dplyr::filter(bad_wind_extent_hu == 1) %>%
          dplyr::mutate(
            wind_extent_warning_id = name %>%
              stringr::str_to_title() %>%
              stringr::str_c(iso_time, sep = ", ")
          ) %>%
          dplyr::select(wind_extent_warning_id) %>%
          purrr::pluck() %>%
          unlist(use.names = F) %>%
          stringr::str_flatten_comma(", and "),
        " are Hurricane observation(s) with missing wind extents. These observation(s) are being dropped. Proceed with Caution."
      )
    )

  data <- data %>%
    dplyr::filter(
      bad_wind_extent_ts != 1  &
        bad_wind_extent_hu != 1
    ) %>%
    dplyr::select(
      -bad_wind_extent_ts,
      -bad_wind_extent_hu
    )

  return(data)
}

#' Converts to a simple tibble
#'
#' Drops the geometry and units, if supplied
#'
#' @param data - input data
#' @return tibble - no units or geometry
#' @export
convert_to_plain_tibble <- function(data) {
  # Removing sf and units
  if ("sf" %in% class(data)){
    data <- data %>%
      sf::st_drop_geometry()
  }
  data <- data %>%
    units::drop_units()

  return(data)
}
