#' Distance Between points
#'
#' Determines the distance from point A to point B
#'
#' @param pt_a the first point
#' @param pt_b the second point
#'
#' @return double or units
calc_distance <- function(pt_a, pt_b) {
  # Coordinates
  coords_a <- sf::st_coordinates(pt_a)
  coords_b <- sf::st_coordinates(pt_b)

  # Distance
  geosphere::distGeo(
    coords_a,
    coords_b
  )
}

calc_bearing <- function(pt_a, pt_b) {
  # Coordinates
  coords_a <- sf::st_coordinates(pt_a)
  coords_b <- sf::st_coordinates(pt_b)

  # Calculating the bearing
  geosphere::bearing(
    coords_a,
    coords_b
  )
}

#' Extracts and Sets the Lat and Lon
#' @param points the points to compare
#' @return tibble with the columns `lat` and `lon` added
#' @export
point2latlon <- function(points) {
  points %>%
    dplyr::mutate(
      coords = sf::st_coordinates(geometry),
      lon = coords[, 1], lat = coords[, 2]
    ) %>%
    dplyr::select(-coords)
}


#' Finds the cyclone interpolated end points
#'
#' This function only interpolates. It does not update any existing variables!
#' The interpolated end points are based upon the point after the cyclone
#' transitions, i.e. the lag is the hurricane point and th current is the
#' tropical storm. The interpolated point is determined via the bearing from
#' the lag to the current, and the distance from the lag to the current.
#'
#' @inheritParams interpolate_points
#' @return sf tibble of the interpolate points where the cyclone ends
#' @export
interpolate_end <- function(data, dist_col, wind_speed, grp_vars = NULL) {
  # Input checking
  check_columns(data, {{dist_col}}, {{grp_vars}}, c("usa_wind", "iso_time", "geometry"))

  # Getting the original names
  orig_names <- colnames(data)

  # Argument Checking
  checkmate::check_class(data, "sf")
  check_columns(data, c("geometry", "iso_time", "usa_wind"), {{ dist_col }}, {{ grp_vars }})

  # Initial setup
  points <- data %>%
    dplyr::group_by(dplyr::across({{ grp_vars }})) %>%
    dplyr::arrange(iso_time, .by_group = TRUE) %>%
    dplyr::mutate(
      dplyr::across(c(usa_wind, iso_time, geometry), .fns = list(lag = dplyr::lag), .names = "{.fn}_{.col}")
    ) %>%
    dplyr::mutate(
      f =  (lag_usa_wind - wind_speed) / (lag_usa_wind - usa_wind)
    ) %>%
    dplyr::mutate(dplyr::across({{ dist_col }}, ~ pmax(
      dplyr::lag(.x) / 2,
      dplyr::lag(.x) * (1-f)
    )))

  # Finding the transition points and filtering
  points <- points %>%
    dplyr::mutate(indicator_flag = lag_usa_wind > wind_speed & usa_wind < wind_speed) %>%
    dplyr::filter(indicator_flag)

  # Setting the timestamp
  points <- points %>%
    dplyr::mutate(
      # Lag is the hu point, currently at TS point
      iso_time = lag_iso_time + (iso_time - lag_iso_time) * f,
      interpolated = 1,
      usa_wind = wind_speed
    ) %>%
    dplyr::ungroup()

  # Updating the geometry
  new_points <- geosphere::destPoint(
      sf::st_coordinates(points$lag_geometry),
      calc_bearing(points$lag_geometry, points$geometry),
      calc_distance(points$geometry, points$lag_geometry) * points$f
    )
  new_geo <- as.data.frame(new_points) %>%
      sf::st_as_sf(coords = c("lon", "lat"), crs = sf::st_crs(points), sf_column_name = "geometry") %>%
      sf::st_as_sfc()
  points <- sf::st_set_geometry(points, new_geo)

  # Updating the lat and long
  points <- point2latlon(points)

  # Removing added columns, if not selected
  points <- points %>%
    dplyr::select(tidyselect::all_of(orig_names), interpolated)

    # Warn for NaNs
  if (any(is.na(dplyr::pull(points, {{ dist_col }})))) {
    rlang::warn("NaNs in interpolation of the `dist_col` in interpolate_end")
  }

  return(points)
}

#' Finds the Cyclone start interpolated points
#'
#' The interpolated start points are based upon the point before the cyclone
#' transitions, i.e. the lead is the hurricane point and the current is the
#' tropical storm. The interpolated point is determined via the bearing from
#' the current to the lead, and the distance from the current to the lead.
#'
#' @inheritParams interpolate_points
#' @return sf tibble of the interpolate points where the cyclone starts
#' @export
interpolate_start <- function(data, dist_col, wind_speed, grp_vars = NULL) {
  # Input checking
  check_columns(data, {{dist_col}}, {{grp_vars}}, c("usa_wind", "iso_time", "geometry"))
  orig_names <- colnames(data)

  # Initial setup
  points <- data %>%
    dplyr::group_by(dplyr::across({{ grp_vars }})) %>%
    dplyr::arrange(iso_time, .by_group = TRUE) %>%
    dplyr::mutate(
      dplyr::across(c(usa_wind, iso_time, geometry), .fns = list(lead = dplyr::lead), .names = "{.fn}_{.col}")
    ) %>%
    dplyr::mutate(f =  1 - ((lead_usa_wind - wind_speed) / (lead_usa_wind - usa_wind))) %>%
    dplyr::mutate(dplyr::across({{ dist_col }}, ~ pmax(
      dplyr::lead(.x) / 2,
      dplyr::lead(.x) * f
    )))

  # Finding the indicator and filtering
  points <- points %>%
    dplyr::mutate(indicator_flag = lead_usa_wind > wind_speed & usa_wind < wind_speed) %>%
    dplyr::filter(indicator_flag)

  # Determining distance to next and timestamp
  points <- points %>%
    dplyr::mutate(
      iso_time = iso_time +
        (lead_iso_time - iso_time) * f,
      interpolated = 1,
      usa_wind = wind_speed
    ) %>%
    dplyr::ungroup()


  # Fixing up the geometry
  new_points <- geosphere::destPoint(
      sf::st_coordinates(points$geometry),
      calc_bearing(points$geometry, points$lead_geometry),
      calc_distance(points$geometry, points$lead_geometry) * points$f
    )
  new_geo <- as.data.frame(new_points) %>%
      sf::st_as_sf(coords = c("lon", "lat"), crs = sf::st_crs(points), sf_column_name = "geometry") %>%
      sf::st_as_sfc()
  points <- sf::st_set_geometry(points, new_geo)

  # Updating the lat and lon
  points <- point2latlon(points)

  # Removing added columns, if not selected
  points <- points %>%
    dplyr::select(tidyselect::all_of(orig_names), interpolated)

  # Warn for NaNs
  if (any(is.na(dplyr::pull(points, {{ dist_col }})))) {
    rlang::warn("NaNs in interpolation of the `dist_col` in interpolate_start")
  }

  return(points)
}

#' Interpolates the start and end
#'
#' Interpolates the start and endpoint of a cyclone. Linear interpolation
#' based upon the `usa_wind` and the `iso_time` column is used to determine
#' the time and the position.
#'
#' @param data input tibble of points. This must contain the columns `usa_wind`,
#' `iso_time`, and `geometry`
#' @param dist_col eg wind_extent_hu_m, interactions_wind_extent_hu_m (or tidyselect)
#' @param wind_speed the wind speed to interpolate to. Typically 34 or 64
#' @param grp_vars  <[`tidy-select`][dplyr_tidy_select]>  of the variables to
#' group on. These typically identifies the storm
#' @return sf tibble of the interpolate points
#' @export
interpolate_points <- function(data, dist_col, wind_speed, grp_vars) {
  checkmate::check_class(data, "sf")
  check_columns(data, {{dist_col}}, {{grp_vars}}, c("usa_wind", "iso_time", "geometry"))

  start <- interpolate_start(data, {{ dist_col }}, wind_speed, {{grp_vars}})
  end <- interpolate_end(data, {{ dist_col }}, wind_speed, {{grp_vars}})

  points <- dplyr::bind_rows(
    data,
    start,
    end
  ) %>%
    dplyr::group_by(dplyr::across({{ grp_vars }})) %>%
    dplyr::arrange(iso_time, .by_group = TRUE) %>%
    dplyr::ungroup() %>%
    dplyr::mutate(interpolated = tidyr::replace_na(interpolated, FALSE))

  return(points)
}

#' Sanity Check on interpolated points
#'
#' Preforms a basic sanity check on the interpolated points to ensure that
#' none of the interpolated points are outside of the interval.
#' The points dataset must contain the columns usa_wind and iso_time.
#'
#' @param points the sf tibble to check
#' @param wind_speed the wind speed that the values should be interpolated to
#' @param grp_vars the grouping variable, typically usa_atcfid or sid
#' @export
check_points <- function(points, wind_speed, grp_vars = NULL) {

  checkmate::expect_class(points, "sf")
  pass <- TRUE

  # Setting up the checks - add on the lag and lead after sorting
  sorted <- points %>%
    dplyr::group_by(dplyr::across({{ grp_vars }})) %>%
    dplyr::arrange(iso_time, .by_group = TRUE) %>%
    point2latlon() %>%
    dplyr::mutate(dplyr::across(.cols = c(iso_time, lat, lon, usa_wind), .fns = c(lead = dplyr::lead, lag = dplyr::lag)))

  # Only checking interpolated rods
  to_check <- sorted %>%
    dplyr::filter(interpolated == TRUE)

  warn_between <- function(valid, col, col_name) {
    to_print <- to_check %>%
      dplyr::filter({{valid}} == FALSE) %>%
      dplyr::select(iso_time, {{col}})
    rlang::warn(glue::glue("{col_name} is not interpolated!"), body = format(to_print))
    return(FALSE)
  }

  # Time should be sorted
  to_check <- to_check %>%
    dplyr::mutate(valid_times = dplyr::between(iso_time, iso_time_lag, iso_time_lead))
  if (!all(to_check$valid_times)){
    pass <- warn_between(valid_times, iso_time, "iso_time")
  }

  # Values should be within end points - interpolation
  check_within <- function(x, a, b) {
    dplyr::between(x, a, b) | dplyr::between(x, b, a)
  }
  to_check <- to_check %>%
    dplyr::mutate(valid_lat = check_within(lat, lat_lead, lat_lag),
                  valid_lon = check_within(lon, lon_lead, lon_lag),
                  valid_wind =  check_within(usa_wind, usa_wind_lead, usa_wind_lag))

  if (!all(to_check$valid_lat)){
    warn_between(valid_lat, lat, "lat")
  }
  if (!all(to_check$valid_lon)){
    pass <- warn_between(valid_lon, lon, "lon")
  }
  if (!all(to_check$valid_wind)){
    pass <- warn_between(valid_wind, usa_wind, "usa_wind")
  }

  if (!all(to_check$usa_wind == wind_speed)){
    rlang::warn(glue::glue("All of the interpolated wind speeds are not {wind_speed}"))
    pass <- FALSE
  }

  return(pass)

}
