# https://github.com/walkerke/tigris/blob/d660228b2027317030b54b6907f3ce340b8c63e4/R/helpers.R

#' Set the cache directory to store shapefiles with hipwits
#'
#' @description By default, hipwits uses the rappdirs package to determine a suitable location to store shapefiles
#' on the user's computer.  However, it is possible that the user would want to store shapefiles in a custom
#' location.  This function allows users to set the cache directory, and stores the result in the user's
#' .Renviron so that hipwits will remember the location.
#'
#' Windows users: please note that you'll need to use double-backslashes or forward slashes
#' when specifying your cache directory's path in R.
#'
#' @param path The full path to the desired cache directory
#' @export
#' @examples \dontrun{
#' # Set the cache directory
#' hipwits_cache_dir("PATH TO MY NEW CACHE DIRECTORY")
#'
#' # Check to see if it has been set correctly
#' Sys.getenv("HIPWITS_CACHE_DIR")
#' }
hipwits_cache_dir <- function(path) {
  home <- Sys.getenv("HOME")
  renv <- file.path(home, ".Renviron")
  if (!file.exists(renv)) {
    file.create(renv)
  }

  check <- readLines(renv)

  if (isTRUE(any(grepl("HIPWITS_CACHE_DIR", check)))) {
    oldenv <- read.table(renv, stringsAsFactors = FALSE)
    newenv <- oldenv[!grepl("HIPWITS_CACHE_DIR", oldenv$V1), ]
    write.table(newenv, renv,
      quote = FALSE, sep = "\n",
      col.names = FALSE, row.names = FALSE
    )
  }

  var <- paste0("HIPWITS_CACHE_DIR=", "'", path, "'")

  write(var, renv, sep = "\n", append = TRUE)
  message(sprintf("Your new hipwits cache directory is %s. \nTo use now, restart R or run `readRenviron('~/.Renviron')`", path))
}

#' Helper function to download a file
#'
#' @param url URL to download
#' @param file_loc location to download to
#' @param refresh Whether to re-download files
#' @param progress_bar If set to FALSE, do not display download progress bar
#'
download_file <- function(url, file_loc, progress_bar, refresh) {
  if (progress_bar) {
    r <- httr::GET(
      url,
      httr::write_disk(file_loc, overwrite = refresh),
      httr::progress(type = "down")
    )
  } else {
    r <- httr::GET(
      url,
      httr::write_disk(file_loc, overwrite = refresh)
    )
  }
  if (r$status_code != 200) {
    rlang::warn(glue::glue("Did not download {url}! Status Code {r$status_code}"))
  }
}

#' Gets the cache directory, building if necessary
get_cache_dir <- function() {
  if (Sys.getenv("HIPWITS_CACHE_DIR") != "") {
    cache_dir <- Sys.getenv("HIPWITS_CACHE_DIR")
    cache_dir <- path.expand(cache_dir)
  } else {
    cache_dir <- rappdirs::user_cache_dir("hipwits")
  }
  if (!file.exists(cache_dir)) {
    dir.create(cache_dir, recursive = TRUE)
  }
  cache_dir
}

#' Helper function to download shapefile data
#'
#' @param url URL for zipped shapefile.
#' @param refresh Whether to re-download shapefiles if cached. Defaults to value of the global
#' option "hipwits_refresh" if that option is, and FALSE if not. This will override the behavior
#' set in "hipwits_refresh" option if a value (TRUE or FALSE) is provided.
#' @param progress_bar If set to FALSE, do not display download progress bar
#' (helpful for R Markdown documents). Defaults to TRUE.
#' @param keep_zipped_shapefile If set to TRUE, do not delete zipped shapefile
#' (stored in temporary directory or HIPWITS_CACHE_DIR depending on the configuration of
#' global option "hipwits_use_cache"). Defaults to FALSE.
#'
#' @return sf or sp data frame
#' @export
load_shapefile <- function(url,
                           refresh = getOption("hipwits_refresh", FALSE),
                           progress_bar = TRUE,
                           keep_zipped_shapefile = FALSE) {
  use_cache <- getOption("hipwits_use_cache", FALSE)

  shape_file <- basename(url)

  obj <- NULL

  if (use_cache) {
    cache_dir <- get_cache_dir()

    if (file.exists(cache_dir)) {
      shape <- gsub(".zip", "", shape_file)
      shape <- gsub("_shp", "", shape) # for historic tracts

      file_loc <- file.path(cache_dir, shape_file)
      shp_loc <- file.path(cache_dir, sprintf("%s.shp", shape))

      if (refresh | !file.exists(shp_loc)) {
        download_file(url, file_loc, progress_bar, refresh)
      }

      unzip_file <- function() {
        unzip(file_loc, exdir = cache_dir, overwrite = TRUE)
      }
      remove_zip <- function() {
        if (file.exists(file_loc) && file.exists(shp_loc)) {
          invisible(file.remove(file_loc))
        }
      }

      if (refresh | !file.exists(shp_loc)) {
        unzip_file()
      } else {
        unzip_file()

        if (!keep_zipped_shapefile) {
          remove_zip()
        }
      }
    }

    obj <- sf::st_read(
      dsn = cache_dir, layer = shape,
      quiet = TRUE, stringsAsFactors = FALSE
    )

    if (is.na(sf::st_crs(obj)$proj4string)) {
      sf::st_crs(obj) <- "+proj=longlat +datum=NAD83 +no_defs"
    }
  } else {
    tmp <- tempdir()
    file_loc <- file.path(tmp, shape_file)

    download_file(url, file_loc, progress_bar, refresh = TRUE)

    obj <- read_shapefile(file_loc)

    if (is.na(sf::st_crs(obj)$proj4string)) {
      sf::st_crs(obj) <- "+proj=longlat +datum=NAD83 +no_defs"
    }
  }
  return(obj)
}


#' Helper function to download lnx data
#'
#' @param keys URL, date, vector of dates, or date range
#' @param use_updated boolean flag to downloaded the "updated" dataset if
#' available
#' @param refresh Whether to re-download lnx if cached. Defaults to value of the global
#' option "hipwits_refresh" if that option is, and FALSE if not. This will override the behavior
#' set in "hipwits_refresh" option if a value (TRUE or FALSE) is provided.
#' @param progress_bar If set to FALSE, do not display download progress bar
#' (helpful for R Markdown documents). Defaults to TRUE.
#'
#' @seealso get_gauge_conus_datasets
#' @return tibble
#' @export
load_lnx <- function(keys,
                     use_updated = TRUE,
                     refresh = getOption("hipwits_refresh", FALSE),
                     progress_bar = TRUE) {
  use_cache <- getOption("hipwits_use_cache", FALSE)



  load_single_lnx <- function(url) {
    lnx_file <- basename(url)

    obj <- NULL

    if (use_cache) {
      cache_dir <- get_cache_dir()

      if (file.exists(cache_dir)) {
        file_loc <- file.path(cache_dir, lnx_file)
        if (refresh | !file.exists(file_loc)) {
          download_file(url, file_loc, progress_bar, refresh)
        }
      }

      # Reading in the file
      obj <- read_lnx_raw(file_loc)
    } else {
      tmp <- tempdir()
      file_loc <- file.path(tmp, lnx_file)

      download_file(url, file_loc, progress_bar, refresh = TRUE)

      # Reading in the file
      obj <- read_lnx_raw(file_loc)
    }

    # https://ftp.cpc.ncep.noaa.gov/precip/CPC_UNI_PRCP/GAUGE_CONUS/DOCU/PRCP_CU_GAUGE_V1.0CONUS_0.25deg.README
    date <- parse_lnx_date(url)
    obj <- dplyr::tibble(
      GRIDID = seq_along(obj),
      rainfall = units::set_units(obj / 10.0, "mm"),
      date = date
    )
    return(obj)
  }

  # Getting the endpoints
  urls <- purrr::map(keys, get_gauge_conus_datasets, use_updated=use_updated)
  results <- purrr::map_dfr(unlist(urls), load_single_lnx)
  return(results)
}
