#' Constructs the GAUGE_CONUS endpoint
#'
#'  https://ftp.cpc.ncep.noaa.gov/precip/CPC_UNI_PRCP/GAUGE_CONUS
#'
#' A check is made when `use_updated` is true to confirm if the update
#' exists
#' @param date date or lubridate like object
#' @param use_updated boolean flag to downloaded the "updated" dataset if
#' available
#' @return endpoint
#' @export
construct_gauge_conus_endpoint <- function(date, use_updated = TRUE) {
  base_endpoint <- getOption("hipwits_gauge_conus_daily_endpoint")

  # constructing the uri
  date <- lubridate::as_date(date)
  year <- lubridate::year(date)

  date_fmt <- format(date, "%Y%m%d")
  fname <- glue::glue("PRCP_CU_GAUGE_V1.0CONUS_0.25deg.lnx.{date_fmt}")

  if (year < 1948) {
    rlang::abort(glue::glue("No data for {year}"))
  } else if (year <= 2006) {
    endpoint <- glue::glue("{base_endpoint}/V1.0/{year}/{fname}.gz")
    return(endpoint)
  } else if (year <= 2018 || !use_updated) {
      if (year <= 2008){
        endpoint <- glue::glue("{base_endpoint}/RT/{year}/{fname}.RT.gz")
      } else {
         endpoint <- glue::glue("{base_endpoint}/RT/{year}/{fname}.RT")
      }
    return(endpoint)
  } else {
    # Check if Updated Exists
    url <- glue::glue("{base_endpoint}/UPDATED/{year}/")
    if (.update_exist(date, url)) {
      endpoint <- glue::glue("{base_endpoint}/UPDATED/{year}/{fname}.UPDATED")
    } else {
      if (year <= 2008){
        endpoint <- glue::glue("{base_endpoint}/RT/{year}/{fname}.RT.gz")
      } else {
         endpoint <- glue::glue("{base_endpoint}/RT/{year}/{fname}.RT")
      }
    }
  }
  return(endpoint)
}

#' Gets the FTP Index for an endpoint (CPC LNX DAILY)
#'
#' This is a utility method and is not exposed in the package
#'
#' @param url url to query
.read_cpc_lnx_ftp_index <- function(url) {
  index <- rvest::read_html(url)

  # Look-ahead and look-behind capture groups
  date_regexp <- getOption("hipwits_gauge_conus_daily_date_regexp")

  content <- index %>%
    rvest::html_element("body") %>%
    rvest::html_table() %>%
    dplyr::mutate(date = stringr::str_extract(Name, date_regexp)) %>%
    dplyr::mutate(date = lubridate::as_date(date)) %>%
    dplyr::mutate(endpoint = glue::glue("{url}{Name}")) # fs::path removes a /

  return(content)
}


#' Determines if the date exists on the url
#'
#' This is a utility method and is not exposed in the package
#'
#' @param date date like object to check
#' @param url url of the ftp index
.update_exist <- function(date, url) {
  index_content <- .read_cpc_lnx_ftp_index(url)
  date <- lubridate::as_date(date)

  return(date %in% index_content$date)
}

#' Builds the endpoint for the index
#'
.construct_gauge_conus_index_endpoint <- function(year, use_updated = TRUE) {
  base_endpoint <- getOption("hipwits_gauge_conus_daily_endpoint")
  checkmate::checkInt(year)

  # constructing the url
  if (year < 1948) {
    rlang::abort(glue::glue("No data for {year}"))
  } else if (year <= 2006) {
    endpoint <- glue::glue("{base_endpoint}/V1.0/{year}/")
  } else if (year <= 2018 || !use_updated) {
    endpoint <- glue::glue("{base_endpoint}/RT/{year}/")
  } else {
    endpoint <- glue::glue("{base_endpoint}/UPDATED/{year}/")
  }
  return(endpoint)
}

#' Constructs the endpoint tibble
#'
#' @param date_range - the range of dates to read in
#' @return tibble of available endpoints
.gauge_conus_available <- function(date_range) {
  checkmate::checkClass(date_range, "Interval")

  # Constructing the years
  dates <- seq(lubridate::int_start(date_range), lubridate::int_end(date_range), by = "1 day")
  years <- purrr::map(dates, lubridate::year) %>%
    unique()

  # Getting the FTP index endpoints
  endpoints <- c(
    purrr::map(years, .construct_gauge_conus_index_endpoint, use_updated = TRUE),
    purrr::map(years, .construct_gauge_conus_index_endpoint, use_updated = FALSE)
  )
  endpoints <- unique(endpoints)

  # Querying all endpoints
  all_endpoints <- purrr::map_dfr(endpoints, .read_cpc_lnx_ftp_index) %>%
    dplyr::filter(!is.na(date))

  return(all_endpoints)
}

#' Downloads CPC_GAUGE_CONUS within a datarange
#'
#' Queries https://ftp.cpc.ncep.noaa.gov/precip/CPC_UNI_PRCP/GAUGE_CONUS for
#' all available precipitation files within the supplied date range and downloads
#' the files to the supplied output directory.
#'
#' @param key URL, date, or date range
#' @param use_updated boolean flag to downloaded the "updated" dataset if
#' available
#' @return list of the endpoints
#' @export
get_gauge_conus_datasets <- function(key, use_updated = TRUE) {

  # Easy case - key is a url
  if (is.character(key) && startsWith(key, "http") ){
    return(key)
  }

  # Finding the date format
  if (lubridate::is.interval(key)) {
    date_range <- key
  } else if ("Date" %in% class(key) || "POSIXct" %in% class(key)) {
    key <- lubridate::as_date(key)
    date_range <- lubridate::interval(key, key)
  } else {
    date_range <- tryCatch(
      expr = {
        lubridate::interval(key)
      },
      error = function(e) {
       rlang::abort(glue::glue("Could not process key {key}. Try passing a url, `lubridate::interval`, or `lubridate::date`"))
      },
      warning = function(w) {
        rlang::warn(glue::glue("Could not process key {key}. Try passing a url, `lubridate::interval`, or `lubridate::date`"))
      }
    )
  }

  # Getting available updates
  checkmate::checkClass(date_range, "Interval")
  endpoints <- .gauge_conus_available(date_range)


  # Filtering for within date range
  endpoints <- endpoints %>%
    dplyr::filter(
      date >= lubridate::date(lubridate::int_start(date_range)),
      date <= lubridate::date(lubridate::int_end(date_range)),
    )

  # Flag for UPDATED and then filter
  endpoints <- endpoints %>%
    dplyr::mutate(is_update = fs::path_ext(Name) == "UPDATED") %>%
    dplyr::arrange(date, is_update)

  if (use_updated) {
    endpoints <- endpoints %>%
      dplyr::group_by(date) %>%
      dplyr::slice_max(order_by = is_update, n = 1) %>%
      dplyr::ungroup()
  } else {
    endpoints <- endpoints %>%
      dplyr::filter(is_update == FALSE)
  }

  endpoints$endpoint

}

#' Read the raw binary Rainfall Data
#'
#' The structure of the fortran file is provided in
#' PRCP_CU_GAUGE_V1.0CONUS_0.25deg.README at
#' https://ftp.cpc.ncep.noaa.gov/precip/CPC_UNI_PRCP/GAUGE_CONUS/DOCU/.
#'
#' A numeric vector is returned, where the index of the vector is the grid
#' cell. The rainfall units are in the default of the CPC, 0.1mm increments.
#'
#' @param fname filename to read
#' @return numeric vector
#' @export
#' @examples \dontrun{
#' # Download the file
#' url <- "https://ftp.cpc.ncep.noaa.gov/precip/CPC_UNI_PRCP/GAUGE_CONUS/RT/2023/PRCP_CU_GAUGE_V1.0CONUS_0.25deg.lnx.20230101.RT"
#' tmp <- tempfile()
#' httr::GET(url, httr::write_disk(tmp))
#'
#' # read the file
#' data <- read_lnx_raw(tmp)
#' }
read_lnx_raw <- function(fname) {

  # Reading in the file
  con <- gzcon(file(fname, "rb"))
  n <- 300*120*2 #  n = 300*120*2 entries, doubled
  obj <- readBin(con, "double", size = 4, n = 300*120, endian = "little")
  close(con)

  # Setting the NA
  obj[obj == -999] <- NA
  obj
}

#' Parses the date from the lnx file name
#'
#' This uses a regular expression to find the date component, and is
#' converted into a lubridate object
#'
#' @param fname the url or path name
#' @return date lubridate date
#' @export
parse_lnx_date <- function(fname) {
    date_regexp <- getOption("hipwits_gauge_conus_daily_date_regexp")
  date_str <- stringr::str_extract(fs::path_file(fname), date_regexp)
  date <- lubridate::as_date(date_str)
  return(date)
}

#' Read the binary Rainfall Data
#'
#' The structure of the fortran file is provided in
#' PRCP_CU_GAUGE_V1.0CONUS_0.25deg.README at
#' https://ftp.cpc.ncep.noaa.gov/precip/CPC_UNI_PRCP/GAUGE_CONUS/DOCU/.
#'
#' @param fname filename to read
#' @param grid the grid to append the rainfall to, or the number of cells. If
#' the number of cells is passed the rate data is returned. If a tibble is
#' passed, then the rainfall is appended.
#' @param na.rm flag to determine to drop na
#' @return vector or tibble. If tibble, then the columns `rainfall` and `date`
#' are appended. The date is parsed from the filename.
#' @seealso read_lnx_raw
#' @export
read_lnx <- function(fname, grid = NULL, na.rm = TRUE){

  # Reading in the fortran style binary file
  rain <- read_lnx_raw(fname)
  rain <- rain / 10.0 # converting from 0.1mm to mm

  # Return if no grid
  if (!inherits(grid, c("tbl", "data.frame"))){
    return(rain)
  }

  # Extracting the date from the file
  date <- parse_lnx_date(fname)



  # Adding to the grid
  precip <- grid %>%
    dplyr::mutate(rainfall = units::set_units(rain, "mm")) %>%
    dplyr::mutate(date = date)

  if (na.rm == TRUE){
    precip <- precip %>%
      dplyr::filter(!is.na(rainfall))
  }

  return(precip)
}
