#' Run Length Encode
#'
#' Encodes a sequence using Run Length Encoding. The returned sequence will be
#' have a sequential number for each continuous run of the same values in the
#' sequence.
#'
#' @seealso rle
#'
#' @param x vector to encode
#'
#' @return encoded sequence
#' @export
#'
#' @examples
#' x <- c("a", "a", "a", "b", "b", "a", "a", "c", "c")
#' rle_encode(x)
rle_encode <- function(x) {
  # Run Length Encode
  r <- rle(x)
  # Ensure values are unique
  r$values <- seq_len(length(r$values))
  # Inverting results
  inverse.rle(r)
}


#' Run Length Encoding Grouping
#'
#' Run length encode the group variables, and then group the data frame.
#' This method first sorts the tibble by the `usa_atcfid` and `iso_time`, and the
#' determines the run length encoding for each of the variables. The tibble is
#' then grouped by the `grp_vars` and the resulting run length encoding vars.
#' This ensures that segments of a track are continuous.
#'
#' @param data ibtracs data
#' @param grp_vars  <[`tidy-select`][dplyr_tidy_select]>  columns to group over
#'
#' @return tibble grouped by the grp_vars and run length encoding group vars.
#'   The run length encoded variables are `ends_with("_rle")`
#' @export
#'
#' @seealso rle_encode
#' @seealso RunLengthEncoding.qmd
rle_group <- function(.data, grp_vars){
  data %>%
    dplyr::ungroup() %>%
    dplyr::arrange(usa_atcfid, iso_time) %>%
    dplyr::mutate(dplyr::across({{grp_vars}}, .fns = rle_encode,.names = "{.col}_rle")) %>%
    dplyr::group_by(dplyr::across({{grp_vars}})) %>%
    dplyr::group_by(dplyr::across(ends_with("_rle")), .add = TRUE)
}
