#' Reads a zip archived shapefile
#'
#' Unzips the archived shapefile and reads in the .shp file. Calls
#' `sf::st_read` to actually import the shapefile. The shapefile is
#' unzipped to a temporary directy.
#'
#' @param fname shapefile file name
#'
#' @return sf tibble
#' @export
read_shapefile <- function(fname){

  # Argument checking
  checkmate::check_file_exists(fname)
  rlang::inform(glue::glue("Reading in the shapefile from {fname}"))

  # Setting up the temporary directory and unzipping
  tmp <- tempdir()

  files <- unzip(
    zipfile = fname,
    exdir = tmp
  )

  # Getting the shapefile and extracting
  shape_file <- files[grepl(pattern = ".shp$", files)]

  if (!fs::file_exists(shape_file)) {
    rlang::warn(
      glue::glue("Can not find a shapefile {shape_file} to read in {tmp}"))
  }

  sf_tibble <- sf::read_sf(shape_file)
  return(sf_tibble)

}

#' Writes data to a shapefile
#'
#' Writes the sf tibble to a zipped shapefile.
#'
#' @param data data to write
#' @param file file or connection to write to. This may include the `.zip` or
#' other extensions, they are stripped with fs::path_ext_remove
#' @param export_dir the directory in which to create the shapefile. If
#' `NULL`, then the directory provided in file is used.
#' @export
write_shapefile <- function(data, file, export_dir = NULL){

  base_name <- fs::path_file(file) %>%
    fs::path_ext_remove()

  if (is.null(export_dir)){
    export_dir <- fs::path_dir(file)
    print(export_dir)
  }

  remove_dir <- FALSE
  zip_dir <- fs::path(export_dir, base_name)
  if (!fs::dir_exists(zip_dir)) {
    remove_dir <- TRUE
    zip_dir <- fs::dir_create(zip_dir)
  }

  # Argument checking
  checkmate::expect_class(data, "sf")

  shp_name <- fs::path_ext_set(base_name, "shp")
  zip_name <- fs::path(export_dir, fs::path_ext_set(base_name, "zip"))

  # append = FALSE overwrites
  sf::st_write(data, fs::path(zip_dir, shp_name), append = FALSE)

  # Creating a zip
  files2zip <- dir(zip_dir, full.names = FALSE)
  zip::zip(zipfile = zip_name, files = files2zip, root = zip_dir)

  # Cleaning up
  if (fs::file_exists(zip_name)) {
    if(remove_dir) {
      fs::dir_delete(zip_dir)
    }
  } else {
    rlang::warn(glue::glue(
      "Could not create a zip! Unzipped files in {zip_dir}"
    ))
  }

  # Return the zipped file name
  invisible(zip_name)

}