#' Determines the Direct Triggers
#'
#' This function intersects the provided corridor with the geometry, and
#' determines the county intersections. It then filters the results to return
#' a unique row for every `id_vars`,`geoid` combination, returning the
#' first time. Note that the date is used for uniqueness, while the time is
#' used to determine the first.
#'
#' @param corridor sf tibble of the corridor
#' @param geo geometry to intersect
#' @param id_vars variable that is used to identify the storm
#' @param county_adj a listing of the adjacent counties
#' @export
wind_triggers <- function(corridor, geo, id_vars = NULL, county_adj = NULL) {

  # Computing the intersection - using intersects instead of intersection as we only care for boolean
  corridor$i <- sf::st_intersects(corridor, geo)
  intersection <- corridor %>%
    sf::st_drop_geometry() %>%
    dplyr::mutate(geoid = purrr::map(i, ~ geo$geoid[.x])) %>%
    tidyr::unnest(geoid)

  # Finding the direct triggers
  triggers <- intersection %>%
    sf::st_drop_geometry() %>%
    dplyr::select(
      {{id_vars}}, iso_time, geoid
    ) %>%
    dplyr::mutate(
      date = lubridate::as_date(iso_time)
    )

  # Filtering for one trigger per storm, keeping the first time
  triggers_unique <- triggers %>%
    dplyr::group_by(
      dplyr::across({{id_vars}}),
      geoid,
    ) %>%
    dplyr::slice_min(n = 1, order_by = iso_time, with_ties =  FALSE) %>%
    dplyr::ungroup()

  # Returning the adjacent, if elected
  if (!is.null(county_adj)) {
    ret_val <- direct_indirect_triggers(triggers_unique,
                                        county_adj = county_adj,
                                        id_vars = {{id_vars}})
  } else{
    ret_val <- triggers_unique
  }

  return(ret_val)
}


#' Direct and Indirect Triggers
#'
#' Takes the direct triggers and finds the indirect triggers through the
#' supplied county adjacency.
#' @param triggers tibble
#' @param county_adj tibble of `geoid` and `adj_geoid`
#' @param id_vars variable that is used to identify the storm
#' @export
direct_indirect_triggers <- function(triggers, county_adj, id_vars) {

  # Crossing the two
  all_triggers <- dplyr::left_join(
    x = triggers,
    y = county_adj,
    by = "geoid",
    multiple = "all"
  )

  # Marking the direct triggers
  all_triggers <- all_triggers %>%
    dplyr::mutate(trigger_type = dplyr::if_else(
        condition = geoid == adj_geoid,
        true =  "Direct",
        false = "Indirect"
      )
    )

  # Setting the triggering geoid, and then renaming
  all_triggers <- all_triggers %>%
    dplyr::rename(
      directly_triggered_geoid = geoid,
      geoid = adj_geoid
    )

  # Checking that when crossing we only introduced one record per triggering cyclone
  # We prefer the "actual" over the indirect
  all_triggers_unique <- all_triggers %>%
      dplyr::group_by(
        dplyr::across({{id_vars}}),
        geoid
      ) %>%
      dplyr::arrange(trigger_type, iso_time) %>%
      dplyr::slice_head(n = 1) %>%
      dplyr::ungroup()

  return(all_triggers_unique)
}

#' Rainfall Triggering Function
#'
#' Determines if a county / region identified by `geoid` has a cumulative
#' rainfall greater than the threshold for the leading and lagging days.
#' The defaults are set according to the RY 2023 Hurricane Data Provisions,
#' https://www.rma.usda.gov/-/media/RMA/Policies/Hurricane-Insurance-Wind-Index/2023/Hurricane-Insurance-Protection-Wind-Index-Endorsement-23-1-HIP-WI/Hurricane-Data-Provisions-2-2023.ashx?la=en
#'
#' The period used for the TS Option is the day the cyclone corridor
#' intersects the county, plus a 1-day lag (prior day) and a 2-day lead
#' (subsequent days), with a threshold of 6 inches.
#'
#' @param data - the rainfall data. Must contain columns `geoid` and `date`
#' @param threshold - the threshold to filter on
#' @param lag_days - the number of lag days to consider
#' @param lead_days - the number of lead days to consider
#' @param default - the default parameter of lead / lag. If not provided,
#' units::as_units("0 mm") is used.
#' @export
rainfall_trigger <- function(data,
    threshold = units::as_units(6, "inch"), lag_days = 1, lead_days = 2,
    default = units::as_units(0, "inch")) {
  lag_days <- checkmate::asCount(lag_days)
  lead_days <- checkmate::asCount(lead_days)

  # Determining the sum
  data <- data %>%
    dplyr::mutate(date = lubridate::as_date(date)) %>%
    dplyr::group_by(geoid) %>%
    dplyr::arrange(date) %>%
    dplyr::mutate(sum_lag = purrr::reduce(purrr::map(1:lag_days, ~ dplyr::lag(rainfall, ., default)), `+`)) %>%
    dplyr::mutate(sum_lead = purrr::reduce(purrr::map(1:lead_days, ~ dplyr::lead(rainfall, ., default)), `+`)) %>%
    dplyr::mutate(cum_sum = sum_lead + sum_lag + rainfall)

  # Filtering
  filtered <- data %>%
    dplyr::filter(cum_sum >= threshold)

  filtered <- filtered %>%
    dplyr::arrange(geoid, date) %>%
    dplyr::mutate(threshold = threshold, lag_days = lag_days, lead_days = lead_days)

  return(filtered)
}


#' Computes the Tropical Storm Wind Index and Rainfall Triggers
#'
#' Combines the tropical storm wind index and rainfall triggers into a
#' common trigger.
#'
#' @param wind_triggers tibble of the wind triggers. This must included the
#' columns `start`, `stop`, and `geoid`
#' @param rainfall_triggers tibble of the rainfall triggers. This must include
#' the columns `geoid`, `date`.
#' @param id the identification parameter to keep to mark the triggering
#' event. If NULL, then usa_atcfid is used.
#' @return tibble of the trigger events. This is a tibble of the `geoid`,
#' `year`, the ID of the
#' @export
#' @importFrom lubridate %within%
wind_rainfall_triggers <- function(wind_triggers, rainfall_trigger, id = NULL){

    # Argument Checking
    check_columns(wind_triggers, cols = c('start', 'stop', 'geoid'))
    check_columns(rainfall_triggers, cols = c('geoid', "date"))

    if (is.null(id)){
      id <- rlang::sym("usa_atcfid")
    }

    # Downselecting to keep only what we need, then computing interval and
    # unnesting the geoid
    wind_triggers <- wind_triggers %>%
      sf::st_drop_geometry() %>%
      dplyr::select(usa_atcfid, start, end, geoid) %>%
      dplyr::mutate(range = lubridate::interval(start, end)) %>%
      tidyr::unnest(geoid)

    rf_wind_triggers <- dplyr::inner_join(
        x = rainfall_triggers %>%
            dplyr::mutate(year = lubridate::year(date)),
        y = wind_triggers %>%
          dplyr::mutate(year = lubridate::year(start)),
        by = c("geoid", "year")
        ) %>%
      dplyr::filter(date %within% range)

    # Finding the number of times per year
    rf_wind_triggers <- rf_wind_triggers %>%
        dplyr::select(geoid, year, {{id}}) %>%
        tidyr::nest(n_ids := c({{id}})) %>%
        dplyr::mutate(n = purrr::map_int(n_ids, dplyr::n_distinct)) %>%
        dplyr::rename("{{id}}s" := n_ids)

    return(rf_wind_triggers)
}
