# Define server logic required to draw a histogram ----
library(shiny)
library(tidyverse)
library(sf)
library(leaflet)
server <- function(input, output) {
  filter_cyclone <- shiny::eventReactive(
    input$selected_storm,
    {
      # Getting the parts from the storm
      storm <- id_season_name %>%
        dplyr::filter(
          storm_name_id == input$selected_storm
          )

      point <- cyclone_points %>%
        dplyr::filter(
          usa_atcfid == storm$usa_atcfid
          )

      county_data <- all_county_data_geo %>%
        dplyr::filter(.data$usa_atcfid == storm$usa_atcfid)

      hu_corridor <- hu_corridor %>%
        dplyr::filter(.data$usa_atcfid == storm$usa_atcfid)

      ts_corridor <- ts_corridor %>%
        dplyr::filter(.data$usa_atcfid == storm$usa_atcfid)


      list(
        point = point,
        county_data = county_data,
        hu_corridor = hu_corridor,
        ts_corridor = ts_corridor
      )
    },
    ignoreNULL = TRUE
  )

  output$rainfall <- DT::renderDataTable(
    {
      filter_cyclone()$county_data %>%
        sf::st_drop_geometry() %>%
        dplyr::select(-storm_name_id,-date,-season, -storm_name,-usa_atcfid) %>%
        dplyr::mutate(dplyr::across(c(name, geoid, statefp, countyfp, trigger_type), forcats::as_factor)) %>%
        DT::datatable(
          filter = "top",
          options = list(
            filter = list(position = "top", clear = TRUE)
          )
        )
      }
    )

  output$leafletMap <- leaflet::renderLeaflet(
    {
    data <- filter_cyclone()

    # AwesomeIcons doesn't seem to work for fa or glpyhicons, which have hurricane
    # icons <- leaflet::awesomeIcons(
    #   icon = "cyclone",
    #   iconColor = "black",
    #   library = "fa",
    #   markerColor = "black"
    # )

    point_labels <- sprintf(
      "<strong>%s</strong><br>(LAT,LON): %3.3f, %3.3f<br>Wind Speed: %s<br>64 Knot Radius [km]: %s<br>34 Knot Radius [km]: %s",
      data$point$iso_time,
      data$point$lat,
      data$point$lon,
      data$point$usa_wind,
      round(data$point$r_64 / 1000, 0),
      round(data$point$r_34 / 1000, 0)
      ) %>%
      lapply(htmltools::HTML)

    county_labels <- sprintf(
      "<strong>%s</strong> <br/> FIP: %s <br/> Date: %s <br/> Trigger Type: %s <br/> Rain Inches: %s",
      data$county_data$name,
      data$county_data$geoid,
      data$county_data$date,
      data$county_data$trigger_type,
      data$county_data$rainfall
      ) %>%
      lapply(htmltools::HTML)

    rainfall_pal <- leaflet::colorBin(
      c('#edf8fb','#bfd3e6','#9ebcda','#8c96c6','#8856a7','#810f7c'),
      domain = data$county_data$rainfall,
      bins = c(0, 4, 5.5, 6, 6.5, 7.5, Inf)
    )

    trigger_pal <- colorFactor(
      c("#a6611a", "#dfc27d", "#018571", "#80cdc1"),
      c("HU Direct", "HU Indirect", "TS Direct", "TS Indirect")
      )


    leaflet() %>%
      addProviderTiles(providers$Stamen.Toner) %>%
      addPolygons(
        data = county_sf,
        fillOpacity = 0,
        color = 'grey',
        weight = 0.5
        ) %>%
      leaflet::addPolygons(
        data = data$county_data,
        fillColor = ~rainfall_pal(rainfall),
        weight = 0,
        fillOpacity = .8,
        label = county_labels,
        labelOptions = labelOptions(
          style = list(
            "font-weight" = "normal",
            padding = "3px 8px"
          ),
          textsize = "15px",
          direction = "auto"
        ),
        group = "Rain"
      ) %>%
      leaflet::addPolygons(
        data = data$ts_corridor,
        color = "#018571",
        opacity = 0,
        fillOpacity =  0.3,
        group = "TS Hulls"
      ) %>%
      leaflet::addPolygons(
        data = data$hu_corridor,
        color = "#a6611a",
        opacity = 0,
        fillOpacity =  0.3,
        group = "HU Hulls"
      ) %>%
      leaflet::addPolygons(
        data = dplyr::filter(
          data$county_data,
          ),
        color = ~trigger_pal(data$county_data$trigger_type),
        weight = 4,
        opacity = 1,
        fillOpacity =  0,
        label = county_labels,
        labelOptions = labelOptions(
          style = list(
            "font-weight" = "normal",
            padding = "3px 8px"
            ),
          textsize = "15px",
          direction = "auto"
          ),
        group = "Triggers"
        ) %>%
  leaflet::addAwesomeMarkers(
    data = data$point,
    # icon = icons,
    label = point_labels,
    group = "Points"
  ) %>%
  leaflet::addLayersControl(
    overlayGroups = c(
      "TS Hulls",
      "HU Hulls",
      "Rain",
      "Triggers",
      "Points"
    ),
    options = leaflet::layersControlOptions(collapsed = FALSE)
  ) %>%
  leaflet::addScaleBar() %>%
  leaflet::hideGroup(c("Points")) %>%
  leaflet::addLegend(
    pal = rainfall_pal,
    values = data$county_data$rainfall,
    opacity = .8,
    title = "Rainfall",
    position = "bottomright"
    ) %>%
      leaflet::addLegend(
        pal = trigger_pal,
        values = data$county_data$trigger_type,
        opacity = 1,
        title = "Trigger Type",
        position = "bottomright"
      )

    }
  )
}
