library(tidyverse)
library(sf)
library(lubridate)

test_that("interpolate end: simple grid a", {
  a <- build_sample_a()

  points <- interpolate_end(a, R, wind_speed = 64)

  # Are we returning the correct type?
  expect_equal(nrow(points), 1)
  expect_s3_class(points, "sf")

  # Are we getting the correct values
  expect_equal(points$usa_wind, 64)
  expect_equal(points$iso_time, lubridate::as_datetime("2020-01-01 06:36:00 UTC"))

  # Lat, Lon
  points <- point2latlon(points)
  expect_equal(unname(points$lon), 0)
  expect_equal(unname(points$lat), 32.2, tolerance = 0.0001)
})

test_that("interpolate end: simple grid b", {
  b <- build_sample_b()

  points <- interpolate_end(b, R, wind_speed = 64)

  # Are we returning the correct type?
  expect_equal(nrow(points), 1)
  expect_s3_class(points, "sf")

  # Are we getting the correct values
  expect_equal(points$usa_wind, 64)
  expect_equal(points$iso_time, lubridate::as_datetime("2020-01-01 06:36:00 UTC"))

  # Lat, Lon
  points <- point2latlon(points)
  expect_equal(unname(points$lon), 22.2, tolerance = 0.001)
  expect_equal(unname(points$lat), 32.2, tolerance = 0.001)
})

test_that("interpolate start: simple grid a", {
  a <- build_sample_a()

  points <- interpolate_start(a, R, wind_speed = 64)

  # Are we returning the correct type?
  expect_equal(nrow(points), 1)
  expect_s3_class(points, "sf")

  # Are we getting the correct values
  expect_equal(points$usa_wind, 64)
  expect_equal(points$iso_time, lubridate::as_datetime("2020-01-01 02:24:00 UTC"), tolerance = lubridate::as.duration("1 min"))

  # Lat, Lon
  points <- point2latlon(points)
  expect_equal(unname(points$lon), 0)
  expect_equal(unname(points$lat), 30.8, tolerance = 0.0001)
})

test_that("interpolate start: simple grid b", {
  b <- build_sample_b()

  points <- interpolate_start(b, R, wind_speed = 64)

  # Are we returning the correct type?
  expect_equal(nrow(points), 1)
  expect_s3_class(points, "sf")

  # Are we getting the correct values
  expect_equal(points$usa_wind, 64)
  expect_equal(points$iso_time, lubridate::as_datetime("2020-01-01 02:24:00 UTC"), tolerance = lubridate::as.duration("1 min"))

  # Lat, Lon
  points <- point2latlon(points)
  expect_equal(unname(points$lon), 20.8, tolerance = 0.01)
  expect_equal(unname(points$lat), 30.8, tolerance = 0.01)
})


test_that("Simple multi-start: c", {
  c <- build_sample_c()

  start <- interpolate_start(c, R, wind_speed = 64)

  # Are we returning the correct type?
  expect_equal(nrow(start), 3)
  expect_s3_class(start, "sf")

  # Are the values correct
  expect_equal(start$usa_wind, c(64, 64, 64))
  expect_equal(start$iso_time,
    c(
      lubridate::as_datetime("2020-01-01 02:24:00 UTC"),
      lubridate::as_datetime("2020-01-01 11:24:00 UTC"),
      lubridate::as_datetime("2020-01-01 18:48:00 UTC")
    ),
    tolerance = lubridate::as.duration("1 min")
  )

  # Lat, Lon
  points <- point2latlon(start)
  expect_equal(unname(points$lon), c(20, 20, 22.26), tolerance = 0.01)
  expect_equal(unname(points$lat), c(30.8, 33.8, 36.27), tolerance = 0.01)

  # Plotting debug
  bind_rows(c, start) %>% arrange(iso_time)
  ggplot() +
    geom_sf(data = c, aes(geometry = geometry), size = 3) +
    geom_sf(data = start, aes(geometry = geometry), size = 3, color = "#ff0000")
})

test_that("Simple multi-end: c", {
  c <- build_sample_c()

  end <- interpolate_end(c, R, wind_speed = 64)

  # Are we returning the correct type?
  expect_equal(nrow(end), 2)
  expect_s3_class(end, "sf")

  # Are the values correct
  expect_equal(end$usa_wind, c(64, 64))
  expect_equal(end$iso_time,
    c(
      lubridate::as_datetime("2020-01-01 06:36:00 UTC"),
      lubridate::as_datetime("2020-01-01 23:12:00 UTC")
    ),
    tolerance = lubridate::as.duration("1 min")
  )

  # Lat, Lon
  # First Point: (65-64)/(65-60) * c(0, 1) + c(20, 32)
  # Second Point: (75-64)/(75-60) * c(2, 1) + c(23, 37)
  points <- point2latlon(end)
  expect_equal(unname(points$lon), c(20, 24.47), tolerance = 0.01)
  expect_equal(unname(points$lat), c(32.2, 37.73), tolerance = 0.01)

  # Plotting debug
  bind_rows(c, end) %>% arrange(iso_time)
  ggplot() +
    geom_sf(data = c, aes(geometry = geometry), size = 3) +
    geom_sf(data = end, aes(geometry = geometry), size = 3, color = "#ff0000")
})


test_that("HDP Example", {
  hdp <- build_sample_hdp()

  hdp_interp <- interpolate_points(hdp, R, wind_speed = 64)

  expect_equal(nrow(hdp_interp), nrow(hdp) +2)
  expect_s3_class(hdp_interp, "sf")

  # Setting up the checks - add on the lag and lead after sorting
  sorted <- hdp_interp %>%
    dplyr::arrange(iso_time) %>%
    point2latlon() %>%
    dplyr::mutate(dplyr::across(.cols = c(iso_time, lat, lon, usa_wind), .fns = c(lead = dplyr::lead, lag = dplyr::lag)))

  # Only checking interpolated rods
  to_check <- sorted %>%
    dplyr::filter(interpolated == TRUE)

  # Time should be sorted
  to_check <- to_check %>%
    mutate(valid_times = dplyr::between(iso_time, iso_time_lag, iso_time_lead))
  expect_true(all(to_check$valid_times))

  # Values should be within end points - interpolation
  check_within <- function(x, a, b) {
    dplyr::between(x, a, b) | dplyr::between(x, b, a)
  }
  to_check <- to_check %>%
    dplyr::mutate(valid_lat = check_within(lat, lat_lead, lat_lag),
                  valid_lon = check_within(lon, lon_lead, lon_lag),
                  valid_wind =  check_within(usa_wind, usa_wind_lead, usa_wind_lag))
  expect_true(all(to_check$valid_lat))
  expect_true(all(to_check$valid_lon))
  expect_true(all(to_check$valid_wind))

  # Interpolated winds should match
  expect_true(all(to_check$usa_wind == 64))

  expect_true(check_points(hdp_interp, wind_speed = 64))

})

test_that("Ian 2022: End", {
  ian <- build_ian() %>%
    filter(lubridate::as_date(iso_time) == '2022-09-30') %>%
    arrange(iso_time) %>%
    slice_tail(n = 3)

  points <- interpolate_end(ian, r_64, wind_speed = 64)

  # Are we returning the correct type?
  expect_equal(nrow(points), 1)
  expect_s3_class(points, "sf")

  # Are we getting the correct values
  expect_equal(points$usa_wind, 64)
  expect_equal(points$iso_time,
               lubridate::as_datetime("2022-09-30 20:32:18 UTC"),
               tolerance = lubridate::as.duration("1 min"))

  # Lat, Lon
  points <- point2latlon(points)
  expect_equal(unname(points$lon), -79.26, tolerance = 0.01)
  expect_equal(unname(points$lat), 33.9, tolerance = 0.01)

  # Plotting debug
  ggplot() +
    geom_sf(data = ian, aes(geometry = geometry), size = 2) +
    geom_sf(data = points, aes(geometry = geometry), size = 3, color = "#ff0000")


})

test_that("Ian 2022 Example", {

  ian <- build_ian()

  ian_interp <- interpolate_points(ian, r_64, wind_speed = 64, grp_vars = usa_atcfid)

  expect_true(check_points(ian_interp, wind_speed = 64))

  ggplot() +
    geom_sf(data = ian, aes(geometry = geometry), size = 3) +
    geom_sf(data = filter(ian_interp, interpolated ==  1),
            aes(geometry = geometry), size = 3, color = "#ff0000")
})

test_that("Ian 2022 Grouping", {

  ian <- build_ian()

  ian_interp <- interpolate_points(ian, r_64, wind_speed = 64, grp_vars = c(usa_atcfid, name))

  expect_true(check_points(ian_interp, wind_speed = 64))

  ggplot() +
    geom_sf(data = ian, aes(geometry = geometry), size = 3) +
    geom_sf(data = filter(ian_interp, interpolated ==  1),
            aes(geometry = geometry), size = 3, color = "#ff0000")
})

test_that("Multi Storm Tibble", {

  storms <- dplyr::bind_rows(
    build_sample_a(),
    build_sample_b(),
    build_sample_c()
  )

  interp <- interpolate_points(storms, R, wind_speed = 64, grp_vars = c(usa_atcfid))

  check_sample <- function(storm_id) {
    storm <- storms %>%
      dplyr::filter(usa_atcfid == storm_id)

    interp_i <- interpolate_points(storm, R, wind_speed = 64, grp_vars = c(usa_atcfid))
    expect_equal(interp_i, dplyr::filter(interp, usa_atcfid == storm_id))
  }
  purrr::map(c("a","b","c"), check_sample)

  expect_true(check_points(interp, wind_speed = 64, grp_vars = c(usa_atcfid)))
})

test_that("Warn on NaN Radii", {

  storms <- build_sample_a() %>%
    mutate(R = NA_real_)

  expect_warning(
    interp <- interpolate_points(storms, R, wind_speed = 64, grp_vars = c(usa_atcfid))
  )

})
